#include <JUI/Widgets/ImageRect.hpp>

#include <Editor/EditorApp.hpp>
#include <Editor/NewMapDialog.hpp>

#include <Format/Format.hpp>

#include "JUI/Widgets/FpsGraph.hpp"
#include <Editor/Grid.hpp>


JUI::Window* EditorApp::CreateAppInfoDialogWindow(JUI::Widget* parent)
{
    // TODO: Implement JUI structure that makes blocks of text easy to impelement.
    auto window = new JUI::Window(parent);
    window->Title("About Editor2D");
    window->Size({350_px, 375_px});
    window->MinSize({350, 375});

    auto* layout = new JUI::VerticalListLayout(window->Content());
    layout->Padding(0_px);

    // TODO: Code like this ends up being a common construct in JUI programs: Make a TextList widget.
    auto line_item = [&] (const std::string& text, int size = 20, const Color4& color = Colors::White) {
        auto* content = new JUI::TextRect(layout);
        content->Size({100_percent, JUI::UDim(size+5, 0)});
        content->Content(text);
        content->TextSize(size);
        content->TextColor(color);
        content->BGColor(Colors::Transparent);
        content->AlignCenterHorizontally();
        content->BorderWidth(0);
    };


    line_item("Redacted Tile Map Editor", 30);
    line_item("Version 1.0", 14);

    auto* box = new JUI::Rect(layout);
    box->Size({100_percent, 200_px});
    box->BGColor(Colors::Transparent);
    box->BorderWidth(0);
    auto* img = new JUI::ImageRect(box);
    img->Content(new JGL::Texture("../icon.png"));
    img->Size({200_px, 200_px});
    img->BGColor(Colors::Transparent);
    img->AnchorPoint({.5f, .5f});
    img->Position({50_percent, 50_percent});
    img->BorderWidth(0);
    img->BGColor(Colors::Transparent);

    line_item("Developed & Maintained by Josh O'Leary.", 16);
    line_item("William R. Maxine H. Ash B.", 12);
    line_item("(c) 2024 - 2025 Redacted Software", 16);

    auto* btn_box = new JUI::Rect(layout);
    btn_box->Size({100_percent, 40_px});
    btn_box->BGColor(Colors::Transparent);
    btn_box->BorderWidth(0);

    auto* btn_layout = new JUI::HorizontalListLayout(btn_box);
    btn_layout->Padding(8_px);

    auto btn_item = [&] (const std::string& text) -> JUI::TextButton* {
        auto* btn = new JUI::TextButton(btn_layout);
        btn->Content(text);
        btn->Size({32_percent, 100_percent});
        btn->TextColor(Colors::Black);
        btn->Center();
        return btn;
    };

    auto* btn_a = btn_item("License");
    auto* btn_b = btn_item("Wiki");
    auto* btn_c = btn_item("Source");

    window->Close();

    return window;
}

Layer * EditorApp::GetLayer(int index) const {
    return loaded_level->layers[index];
}

Vector2i EditorApp::IndexToCell(int index, int width)
{
    int x = index % width;
    int y = index / width;
    return {x, y};
}

void EditorApp::ReplaceFill(int x, int y, int tileid)
{

    std::vector<Vector2i> matching_adjacent_cells;

    int replace = GetTile(x, y);
    SetTile(x, y, tileid);

    int range = 1;

    bool exhausted = false;

    while (!exhausted)
    {
        int stopped = 0;
        if (GetTile(x-range, y) == replace)
            matching_adjacent_cells.push_back({x-range, y});
        else
            stopped++;
        if (GetTile(x+range, y) == replace)
            matching_adjacent_cells.push_back({x+range, y});
        else
            stopped++;
        if (GetTile(x, y-range) == replace)
            matching_adjacent_cells.push_back({x, y-range});
        else
            stopped++;
        if (GetTile(x, y+range) == replace)
            matching_adjacent_cells.push_back({x, y+range});
        else
            stopped++;

        if (stopped == 4)
            exhausted = true;

        range++;
    }


    for (auto& coords : matching_adjacent_cells)
        SetTile(coords, tileid);
}

void EditorApp::FloodFill(int x, int y, int tileid)
{

    if (GetTile(x, y) == -1)
    {
        SetTile(x, y, tileid);
        if (GetTile(x-1, y) == -1) FloodFill(x-1, y, tileid);
        if (GetTile(x+1, y) == -1) FloodFill(x+1, y, tileid);
        if (GetTile(x, y-1) == -1) FloodFill(x, y-1, tileid);
        if (GetTile(x, y+1) == -1) FloodFill(x, y+1, tileid);
    }
}

int EditorApp::CellToIndex(Vector2i cell, int width)
{
    return cell.y*width + cell.x;
}


EditorApp::EditorApp(int width, int height) : OpenGLWindow("Editor App", width, height, GL_VER_MAJOR, GL_VER_MINOR)
{
    camera.DefaultState();
}

EditorApp::EditorApp(): EditorApp(1024, 600) {}

void EditorApp::LoadPreferences()
{
	if (std::filesystem::exists("preferences.json"))
	{

		auto text = read_file_contents("preferences.json");

		if (text.empty())
			return;

		auto [data , errcode] = json::parse(text);

		if (!errcode.empty())
		{
			std::cerr << "Problem!1!" << std::endl;
			return;
		}

		preferences = Preferences(data);
		camera.translation.goal = preferences.recent_camera_state.translation;
		camera.rotation.goal = preferences.recent_camera_state.rotation;
		camera.scale.goal = preferences.recent_camera_state.scale;
	}
}

void EditorApp::SavePreferences()
{

	preferences.recent_camera_state.translation = camera.translation.goal;
	preferences.recent_camera_state.rotation = camera.rotation.goal;
	preferences.recent_camera_state.scale = camera.scale.goal;

	write_file_contents("preferences.json", json::deparse(preferences.Serialize()));

}


void EditorApp::LoadTestFile()
{
    if (std::filesystem::exists("level.json"))
        LoadLevel("level.json");
    else
        CreateTestLevel();
}

void EditorApp::SaveCurrentLevel() const
{
    // TODO: Is it decent for the layers to be written to file separate?
    for (auto layer : loaded_level->layers)
    {
        layer->Save();
    }

    // save current timestamp to level file.
    loaded_level->last_edit_timestamp = std::format("{:%d-%m-%Y %H:%M:%OS}", std::chrono::system_clock::now());

    auto data = loaded_level->Serialize();

    write_file_contents("level.json", json::deparse(data));
}

void EditorApp::SaveTileset() const {
    write_file_contents("tileset.json", json::deparse(loaded_tileset->Serialize()));
}

void EditorApp::SaveCurrentLevelAs(const std::string& filename) const {

    // TODO: Is it decent for the layers to be written to file separate?
    for (auto layer : loaded_level->layers)
    {
        layer->Save();
    }


    auto data = loaded_level->Serialize();


    write_file_contents(filename, json::deparse(data));
}

void EditorApp::LoadLevel(const std::filesystem::path& level_meta_path)
{
    loaded_level = new Level(level_meta_path);

    loaded_tileset = new Tileset(std::filesystem::path(loaded_level->tileset_path));

    loaded_tilesheet = new JGL::Texture(loaded_tileset->texture_path, FilteringMode::NEAREST);

    for (auto layer : loaded_level->layers)
        layer->Load();

    // Check layers for tiles that are too large
    for (auto layer : loaded_level->layers) {
        for (int x = 0; x < layer->rows; x++) {
            for (int y = 0; y < layer->cols; y++) {
                int id = layer->cells[x][y];

                if (id >= loaded_tileset->tile_count)
                    layer->cells[x][y] = -1;
            }
        }
    }

    tileset_view = new TilesetView(loaded_tileset, loaded_tilesheet, scene);

    tileset_view->TileSelected += [this](int id) mutable {
        selected_quad = id;
    };

    tileset_view->TileCollidesChanged += [this](int id, bool collides) mutable {
        loaded_tileset->tiles[id].collides = collides;
        loaded_tileset->tiles[id].persistent = true;
        SaveTileset();
    };

    tileset_view->TileNameChanged += [this](int id, const std::string& new_name) mutable {
        loaded_tileset->tiles[id].name = new_name;
        loaded_tileset->tiles[id].persistent = true;
        SaveTileset();
    };

    layer_view->UpdateComponents(loaded_level);
    layer_view->Open();

    data_ready = true;
}

void EditorApp::CreateTestLevel() {
    loaded_level = new Level();
    loaded_level->name = "TestLevel";
    loaded_level->author = "dawsh";
    loaded_level->cols = 64;
    loaded_level->rows = 64;
    loaded_level->tileset_path = "tileset.json";
    auto layer = new Layer();
    layer->name = "Layer 1";
    layer->cols = 64;
    layer->rows = 64;
    layer->cell_height = 16;
    layer->cell_width = 16;
    layer->binary_path = "layers/1.bin";
    layer->InitGrid();
    layer->Save();
    loaded_level->layers.push_back(layer);
    loaded_tileset = new Tileset();
    loaded_tileset->name = "MegaCommando";
    loaded_tileset->author = "dawsh";
    loaded_tileset->tex_height = 800;
    loaded_tileset->tex_width = 1024;
    loaded_tileset->tile_width = 16;
    loaded_tileset->tile_height = 16;
    loaded_tileset->tile_spacing = 0;
    loaded_tileset->cols = 50;
    loaded_tileset->rows = 64;
    loaded_tileset->file_path = "tileset.json";
    loaded_tileset->texture_path = "assets/megacommando.png";
    loaded_tilesheet = new JGL::Texture("assets/megacommando.png");

    write_file_contents("tileset.json", json::deparse(loaded_tileset->Serialize()));

    //layer_view->UpdateComponents(loaded_level);

    data_ready = true;

    SaveCurrentLevel();
}


void EditorApp::SizeCellIndicatorToTilesetSize()
{
    cell_indicator->Size(JUI::UDim2::FromPixels(loaded_tileset->tile_width, loaded_tileset->tile_height));
}

JUI::Window* EditorApp::CreateTilesetViewerWindow(JUI::Widget* parent)
{
    using namespace JUI;

    auto* wind = new Window(parent);

    return new TilesetView(loaded_tileset, loaded_tilesheet);

    /*wind->Title("Tileset Viewer");
    wind->Size(JUI::UDim2::FromPixels(loaded_tilesheet->GetDimensions().x, loaded_tilesheet->GetDimensions().y+20));
    wind->SetResizable(false);

    wind->Content()->BGColor(Colors::Black);

    auto* img = new Image(wind->Content());
    img->Content(loaded_tilesheet);
    img->FitImageToParent(false);

    auto* grid_overlay = new JGL::RenderTarget(loaded_tilesheet->GetDimensions());
    // TODO: supplying a translucent grid_overlay_color seems to not work?
    J2D::Begin(grid_overlay);
    DrawGrid(AABB2D({0,0}, Vector2(loaded_tilesheet->GetDimensions())), grid_overlay_color);
    J2D::End();

    auto* grid_overlay_tex = new JGL::Texture(*grid_overlay->GetTexture());

    auto* overlay = new JUI::Image(wind->Content());
    overlay->Content(grid_overlay_tex);
    overlay->ZIndex(2);

    cell_indicator = new JUI::Rect(wind->Content());
    cell_indicator->BorderMode(JUI::BorderMode::Outline);
    cell_indicator->BGColor(Colors::Transparent);
    cell_indicator->BorderColor(Colors::Blues::CornflowerBlue);
    cell_indicator->BorderWidth(2);

    SizeCellIndicatorToTilesetSize();*/


    return wind;
}

namespace misc {
    std::vector<std::string> string_expand(const std::string& input, char delimiter = ' ');
    bool string_matches(const std::string& x, const std::vector<std::string>& v);

    std::vector<std::string> string_expand(const std::string& input, char delimiter)
    {
        std::vector<std::string> result;
        std::stringstream ss (input);
        std::string item;

        while (getline(ss, item, delimiter)) {
            result.push_back(item);
        }
        return result;

    }

    auto string_matches(const std::string &x, const std::vector<std::string> &v) -> bool {

        return std::find(v.begin(), v.end(), x) != v.end();

    }

}

struct Arg {
    std::string name;
    std::string type;
    bool optional;
};

struct Command {
    std::string name;
    std::vector<std::string> aliases;
    std::vector<Arg> args;
    std::string err_msg = "ERROR: Invalid command syntax!";
};

void EditorApp::LogHelpInfo() {
    console->Log("Redacted Software 2D Level Editor - Command Line");
    console->Log("Type `cmds` to see a list of all commands!");
}

std::string tolower(const std::string& s) {
    std::string copy = "";
    for (auto elem : s) {
        copy += std::tolower(elem);
    }
    return copy;
}


void EditorApp::BindConsoleCallbacks()
{
    // TODO: This parsing pattern is duplicated between at least 2 other projects at this point.
    // TODO: Move up into JUI or a separate package.
    console->OnInput += [this] (const std::string& input)
    {
        auto tokens = split(input, ' ');

        std::string cmd = tokens[0];

        tokens.erase(tokens.begin());

        if (cmd == "help") {
            return LogHelpInfo();
        }
        if (cmd == "cmds" || cmd == "commandlist") {

        }
        if (cmd == "q" || cmd == "quit") { return this->Close(); }
        if (cmd == "framegraph") {
            // TODO: Add FpsGraph::Toggle().
            fps_graph->Visible(!fps_graph->IsVisible());
        }
        if (cmd == "load") {
            if (tokens.size() >= 1) {
                return LoadLevel(tokens[0]);
            }
           return console->Log("ERROR: Must provide file name!");
        }
        if (cmd == "save") {
            SaveCurrentLevel();
        }
        if (cmd == "save-as") {
            if (tokens.size() >= 1)
                return SaveCurrentLevelAs(tokens[0]);
            else
                return console->Log(std::format("ERROR: Must provide file name: {}", cmd));
        }
        if (cmd == "new") {

        }

        if (cmd == "info") {

        }
        if (cmd == "layer") {

        }
        if (cmd == "tileset") {

        }
        console->Log(std::format("No such command: {}", cmd), Colors::Red);
    };
}


void EditorApp::CreateTopbarWidgets() {
    // Topbar Creation
    topbar = new JUI::UtilityBar(scene);

    topbar_stats = new JUI::Text(topbar);
    topbar_stats->HorizontalTextAlign(JUI::TextAlign::H::Right);
    topbar_stats->TextColor(Colors::Black);

    auto* file = topbar->AddSubmenu("File");
    file->Font(JGL::Fonts::Jupiteroid);

    auto* new_map = file->AddButton("New Map", [this] () mutable {
        nmd->Open();
    });

    file->AddButton("Open", [this]() mutable {

        if (file_dialog) {
            //file_dialog->Close();
            //file_dialog->Parent(nullptr);
            //delete file_dialog;
        }


        file_dialog = new JUI::FileDialogWindow(scene, ".");
    });
    file->AddButton("Save", [this]{SaveCurrentLevel();});
    file->AddButton("Save As");
    file->AddSeparator(2_px);
    file->AddButton("About", [this]{app_info_dialog->Toggle(); });
    file->AddButton("Preferences");

    auto* edit = topbar->AddSubmenu("Edit");
    edit->AddButton("Undo");
    edit->AddButton("Redo");
    edit->AddButton("Copy");
    edit->AddSeparator(2_px);
    edit->AddButton("Paste");
    edit->AddButton("Cut Selection");

    auto* view = topbar->AddSubmenu("View");
    view->AddButton("Zoom In");
    view->AddButton("Zoom Out");
    view->AddSeparator(2_px);
    view->AddButton("Toggle Grid", [this]{ToggleGrid();});
    view->AddButton("Set Background Color", [this]{bg_color_tool_window->Toggle();});
    view->AddSeparator(2_px);
    view->AddButton("Console", [this]()mutable { console->Toggle();});
    view->AddButton("Tileset Viewer", [this]()mutable { tileset_view->Toggle();});

    auto* level = topbar->AddSubmenu("Level");
    auto* layer = topbar->AddSubmenu("Layer");
    layer->AddButton("New");
    layer->AddButton("Open from File");
    layer->AddButton("Duplicate Selected");
    layer->AddButton("Delete Selected");
    layer->AddButton("Edit Selected");
    layer->AddButton("Export Layer");
}

void EditorApp::CreateWidgets()
{
    scene = new JUI::Scene();


    /*auto* blank_text = new JUI::TextRect(scene);
    blank_text->Content("Open or create a new level file to get started!");
    blank_text->Center();
    blank_text->AutoFitSizeToText(true);
    blank_text->Position({20_px, 20_px});*/



    app_info_dialog = CreateAppInfoDialogWindow(scene);

    fps_graph = new JUI::FpsGraph(scene);
    // TODO: Revise to use PseudoDockedElementAtBottomOfViewport coming in the next JUI release.
    fps_graph->Size({100_percent, 50_px});
    fps_graph->AnchorPoint({1, 1});
    fps_graph->Position({100_percent, 100_percent});
    fps_graph->BGColor(Colors::Transparent);
    fps_graph->BorderColor(Colors::Transparent);
    fps_graph->ZIndex(10);

	console = new JUI::CommandLine(scene);
	console->Close();
    BindConsoleCallbacks();

	nmd = new NewMapDialog(scene);
	nmd->Close();

    layer_view = new LayerView(scene);
    layer_view->Close();

    CreateTopbarWidgets();


    bg_color_tool_window = new JUI::Window(scene);
    bg_color_tool_window->Close();

    bg_color_tool = new JUI::ColorPicker(bg_color_tool_window->Content());
    bg_color_tool->OnColorValueChanged += [this] (Color4 new_color) mutable { bg_color = new_color; };

}

void EditorApp::ToggleGrid()
{
    grid_overlay_enabled = !grid_overlay_enabled;
}


int EditorApp::GetLayerIndexFromName(const std::string& name) {

    for (int i = 0; i < loaded_level->layers.size(); ++i) {
        if (loaded_level->layers[i]->name == name) {
            return i;
        }
    }

    return -1;
}

bool EditorApp::Open()
{
	LoadPreferences();

    if (!OpenGLWindow::Open()) return false;

    auto size = GetSize();
    auto vec_size = Vector2i(size.x, size.y);
    if (!JGL::Init(vec_size, 0.f, 1.f))
        return false;
    JGL::Update(vec_size);

    glClearColor(0.f, 0.f, 0.f, 0.f);

    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);
    glDepthMask(GL_TRUE);
    // TODO: Delete when we update to the next release of JGL
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1); // NOTE: This MUST be called for text rendering to work properly!!!

    // TODO: More sophisticated order-of-operations is required.
    // 1. Initialize elements of widgets that are independent of the level itself / the level depends on.
    // 2. Initialize the level.
    // 3. Initialize widgets that are dependent on the level itself.

    //LoadTestFile();

    CreateWidgets();

    //layer_view->UpdateComponents(loaded_level);
    layer_view->ActiveLayerSelected += [this](const std::string& id) mutable {

        int layer_index = GetLayerIndexFromName(id);
        focus_layer_index = layer_index;
    };
    layer_view->LayerVisibilityToggled += [this](const std::string& id, bool value) mutable {
        std::cout << id << ": " << value << std::endl;

        int layer_index = GetLayerIndexFromName(id);
        loaded_level->layers[layer_index]->visible = value;
    };

    /*for (auto tile : loaded_tileset->tiles) {
        std::cout << tile.id << ", " << tile.name << std::endl;
    }*/

    return true;
}

Vector2 EditorApp::GetMouseV2() const {
    auto maus = GetMouseCoordinates();

    return Vector2(maus.x, maus.y);
}
Vector2i EditorApp::GetMouseV2i() const {
    auto maus = GetMouseCoordinates();
    return Vector2i(maus.x, maus.y);
}

Vector2i EditorApp::GetGridCellFromMouse()
{
    auto maus = GetMouseCoordinates();

    Vector2 mouse_v2(maus.x, maus.y);


    mouse_v2 = camera.ScreenToCameraSpace(mouse_v2);

    return Vector2i(
        Math::Floor(mouse_v2.x/GetFocusLayer()->cell_width),
        Math::Floor(mouse_v2.y/GetFocusLayer()->cell_height));
}

void EditorApp::CameraUpdate(float elapsed)
{
    float move_rate = 120;
    float zoom_rate = 0.2f;

    if (IsKeyDown(Keys::LeftArrow))
        camera.translation.goal.x -= move_rate*elapsed;
    if (IsKeyDown(Keys::RightArrow))
        camera.translation.goal.x += move_rate*elapsed;
    if (IsKeyDown(Keys::UpArrow))
        camera.translation.goal.y -= move_rate*elapsed;
    if (IsKeyDown(Keys::DownArrow))
        camera.translation.goal.y += move_rate*elapsed;

    if (IsKeyDown(Keys::Minus))
        camera.scale.goal -= zoom_rate*elapsed;
    if (IsKeyDown(Keys::Equals))
        camera.scale.goal += zoom_rate*elapsed;

    camera.Update(elapsed);

}

int EditorApp::GetTile(const Vector2i& cell)
{
    if (cell.x < 0) return -2;
    if (cell.y < 0) return -2;

    // out of bounds horizontally
    if (cell.x >= GetFocusLayer()->rows) return -2;

    // out of bounds vertically
    if (cell.y >= GetFocusLayer()->cols) return -2;

    return GetFocusLayer()->cells[cell.x][cell.y];
}

int EditorApp::GetTile(int x, int y) { return GetTile({x, y}); }
void EditorApp::SetTile(int x, int y, int tileID) { SetTile({x, y}, tileID);}

void EditorApp::SetTile(const Vector2i& cell, int tileID)
{

	if (cell.x < 0) return;

	if (cell.y < 0) return;

    // out of bounds horizontally
    if (cell.x >= GetFocusLayer()->rows) return;

    // out of bounds vertically
    if (cell.y >= GetFocusLayer()->cols) return;

    GetFocusLayer()->cells[cell.x][cell.y] = tileID;
}

void EditorApp::PlaceTileBrush(const Vector2i& cell)
{

	if (cell.x < 0) return;

	if (cell.y < 0) return;

    // out of bounds horizontally
    if (cell.x > GetFocusLayer()->rows) return;

    // out of bounds vertically
    if (cell.y > GetFocusLayer()->cols) return;


    if (IsKeyDown(Keys::B))
        FloodFill(cell.x, cell.y, selected_quad);
    else if (IsKeyDown(Keys::R))
        ReplaceFill(cell.x, cell.y, selected_quad);
    else
        SetTile(cell, selected_quad);
}

void EditorApp::EraseTile(const Vector2i& cell)
{
	if (cell.x < 0) return;

	if (cell.y < 0) return;

    // out of bounds horizontally
    if (cell.x > GetFocusLayer()->rows) return;

    // out of bounds vertically
    if (cell.y > GetFocusLayer()->cols) return;

    SetTile(cell, -1);
}

bool EditorApp::IsMouseFocusedAnyActiveWidget()
{

    if (bg_color_tool->IsVisible() && bg_color_tool->IsMouseInside())
        return true;

    if (layer_view->IsVisible() && layer_view->IsMouseInside())
        return true;

    if (nmd->IsVisible() && nmd->IsMouseInside())
        return true;

    if (console->IsVisible() && console->IsMouseInside())
        return true;

    if (console->IsVisible() && console->IsMouseInside())
        return true;

    if (tileset_view != nullptr && tileset_view->IsVisible() && tileset_view->IsMouseInside())
        return true;

    if (topbar->IsVisible() && topbar->IsMouseInsideDescendants(true))
        return true;


    //for (auto child : scene->GetDescendants())
        //if (child->GetParent()->IsVisible() && child->IsVisible() && child->IsMouseInside())
            //return true;


    return false;
}

void EditorApp::Update(float elapsed) {
    CameraUpdate(elapsed);
    auto size = GetSize();
    Vector2i vSize = Vector2i(size.x, size.y);

    camera.screenSize.x = vSize.x;
    camera.screenSize.y = vSize.y;

    JGL::Update(vSize);
    scene->SetViewportSize(Vector2(vSize));
    scene->Update(elapsed);


    // TODO: Widget::IsMouseInsideVisibleDescendants(); I.e. ShouldMouseInputsPayAttention?
    //bool mouse_hovering_ui_element = scene->IsMouseInsideDescendants();

    bool mouse_hovering_ui_element = IsMouseFocusedAnyActiveWidget();
    //for (auto child : scene->GetChildren())
    //    if (child->IsVisible() && child->IsMouseInsideDescendants(true))
    //        mouse_hovering_ui_element = true;


    if (!mouse_hovering_ui_element) {
        if (data_ready) {
            if (IsMouseButtonDown(MouseButtons::Left))
                PlaceTileBrush(GetGridCellFromMouse());

            if (IsMouseButtonDown(MouseButtons::Right))
                EraseTile(GetGridCellFromMouse());
        }
    }
}

void EditorApp::DrawGrid(const AABB2D& bounds, const Color4& color)
{
    Vector2 cell_size (loaded_tileset->tile_width, loaded_tileset->tile_height);
    DrawGrid2(bounds, cell_size, color);
}

void EditorApp::DrawGrid(const Color4& color)
{

    auto grid_rows = GetFocusLayer()->rows;
    auto grid_cols = GetFocusLayer()->cols;
    auto grid_pixel_width = GetFocusLayer()->cell_width;
    auto grid_pixel_height = GetFocusLayer()->cell_height;

    float sw = GetWidth();
    float sh = GetHeight();

    AABB2D viewport = {{0,0}, {static_cast<float>(grid_rows*grid_pixel_width), static_cast<float>(grid_cols*grid_pixel_height)}};

    Vector2 cell_size (loaded_tileset->tile_width, loaded_tileset->tile_height);

    DrawGrid(viewport, color);
}

void EditorApp::DrawCellPointerOutline()
{
    Vector2 rel_mouse = Vector2(GetGridCellFromMouse());

    rel_mouse.x *= GetFocusLayer()->cell_width;
    rel_mouse.y *= GetFocusLayer()->cell_height;

    J2D::OutlineRect(cell_pointer_outline_color, rel_mouse, Vector2( GetFocusLayer()->cell_width, GetFocusLayer()->cell_height), cell_pointer_outline_width);
}

void EditorApp::DrawTiles()
{
    /*for (int gx = 0; gx < grid_rows; gx++) {
        for (int gy = 0; gy < grid_cols; gy++) {

            auto quad_idx = grid[gx][gy];

            Vector2 pos(gx*grid_pixel_width, gy*grid_pixel_height);

            //if ((gx+gy) % 2 == 0)
                //J2D::FillRect(Colors::Purples::Fuchsia, pos, {16, 16});

            if (quad_idx < 0)
                continue;

            //auto quad = quads[quad_idx];

            //J2D::DrawPartialSprite(test_tilesheet, pos,
            //                       Vector2(quad.x, quad.y), Vector2(quad.w, quad.h));
        }
    }*/
}

void EditorApp::DrawLayer(const Layer* layer) const
{
    if (!layer->visible) return;

    for (int gx = 0; gx < layer->rows; gx++) {
        for (int gy = 0; gy < layer->cols; gy++) {

            auto quad_idx = layer->cells[gx][gy];

            Vector2 pos(gx*layer->cell_width, gy*layer->cell_height);

            //if ((gx+gy) % 2 == 0)
            //J2D::FillRect(Colors::Purples::Fuchsia, pos, {16, 16});

            if (quad_idx < 0)
                continue;

            auto quad = loaded_tileset->quads[quad_idx];

            J2D::DrawPartialSprite(loaded_tilesheet, pos,
                                   Vector2(quad.x, quad.y), Vector2(quad.w, quad.h));
        }
    }
}

void EditorApp::DrawLevel(const Level* level) const
{
    // TODO: Draw Per-Level Background Texture.

    // TODO: Draw Per-Layer Background Texture.
    for (const auto* layer : level->layers)
    {
        DrawLayer(layer);
    }

    // TODO: Support a handful of presets to represent entity shape types.
    for (auto entity : level->entities)
    {
        J2D::DrawPoint(entity->overlay_color, entity->x, entity->y, 1);
        J2D::DrawString(Colors::White, std::format("{} - {}", entity->type, entity->name), entity->x, entity->y, 1.f, 12);
    }
}

bool EditorApp::IsMouseGridCellInBounds() {
    Vector2i cell = GetGridCellFromMouse();

    if (cell.x < 0) return false;
    if (cell.y < 0) return false;

    if (cell.x >= loaded_level->rows) return false;
    if (cell.y >= loaded_level->cols) return false;

    return true;
}

void EditorApp::Draw()
{
    glClearColor(bg_color.RN(), bg_color.GN(), bg_color.BN(), 1);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    J2D::Begin();
    glPushMatrix();
    glTranslatef(-camera.translation.current.x, -camera.translation.current.y, 0);
    glRotatef(camera.rotation.current, 0, 0, 1);
    glScalef(camera.scale.current, camera.scale.current, 1);




    if (data_ready) {
        if (grid_overlay_enabled)
            DrawGrid(grid_overlay_color);
        DrawLevel(loaded_level);
        if (grid_overlay_enabled)
            DrawGrid(grid_overlay_color);

        DrawCellPointerOutline();
    }


    glPopMatrix();

    if (data_ready) {
        auto maus = GetMouseCoordinates();

        Vector2 mouse_v2(maus.x, maus.y);

        auto bruhbruh = camera.CameraToScreenSpace(camera.ScreenToCameraSpace(mouse_v2));

        if (selected_quad > -1){
            Tile selected_tile = loaded_tileset->tiles[selected_quad];
            J2D::DrawString(Colors::White, std::format("Selected Tile: {}", selected_tile.name), 16, 48, 16);

        }

        Vector2i cell = GetGridCellFromMouse();

        int hovered_tile_id = GetTile(cell);

        if (hovered_tile_id > -1) {
            Tile hovered_tile = loaded_tileset->tiles[hovered_tile_id];
            J2D::DrawString(Colors::White, std::format("Hovered Tile: {}", hovered_tile.name), 16, 64, 16);
        }

        J2D::DrawPoint(Colors::White, bruhbruh, 4);

    }
    J2D::End();

    scene->Draw();
}

void EditorApp::OnClosing()
{
	SavePreferences();
    SaveCurrentLevel();
}

void EditorApp::OnRefresh(float elapsed)
{
    Update(elapsed);
    Draw();
    SwapBuffers();
}

JUI::MouseButton EditorApp::ToJUIEnum(const MouseButton& btn)
{
    if (btn == MouseButtons::Left)   return JUI::MouseButton::Left;
    if (btn == MouseButtons::Middle) return JUI::MouseButton::Middle;
    if (btn == MouseButtons::Right)  return JUI::MouseButton::Right;

    // Default condition.
    return JUI::MouseButton::Left;
}

void EditorApp::TilePickAction() {
    Vector2i cell = GetGridCellFromMouse();

    selected_quad = GetFocusLayer()->cells[cell.x][cell.y];
}

void EditorApp::OnMouseButtonDown(const MouseButtonDownEvent& e)
{
    auto btn = ToJUIEnum(e.Button);

    if (scene->ObserveMouseInput(btn, true)) return;

    // Tile Pick - Middle Click
    if (btn == JUI::MouseButton::Middle)
    {
        return TilePickAction();
    }
}

void EditorApp::OnMouseButtonUp(const MouseButtonUpEvent& e)
{
    auto btn = ToJUIEnum(e.Button);
    if (scene->ObserveMouseInput(btn, false)) return;

    /*if (tileset_viewer->Content()->IsMouseInside()) {
        Vector2i cell = GetTilesetCellFromMouse();
        int index = CellToIndex(cell, loaded_tileset->rows);
        selected_quad = index;

    }*/
    // TODO: This always runs even if we release the mouse inside the window. ObserveMouseInput should handle the case.
}

void EditorApp::OnMouseMove(const MouseMoveEvent& e)
{
    Vector2 mposv2(e.Position.x, e.Position.y);
    if (scene->ObserveMouseMovement(mposv2)) return;

}

void EditorApp::OnKeyDown(const KeyDownEvent& e)
{
    if (scene->ObserveKeyInput(e.key, true)) return;
}

void EditorApp::OnKeyUp(const KeyUpEvent& e)
{


    if (scene->ObserveKeyInput(e.key, false)) return;

    if (e.key == Keys::T && tileset_view) {
        tileset_view->Toggle();
    }

}
