#include <Format/Level.hpp>
#include <Colors.hpp>


Level::Level(): rows(0), cols(0)
{
}

Level::Level(const std::filesystem::path& path)
{
	if (path.empty())
		throw std::runtime_error(std::format("Level::Level({}: blank file path provided!", path.string()));;
	if (!std::filesystem::exists(path))
		throw std::runtime_error(std::format("Level::Level({}): file does not exist!", path.string()));

	std::string file_contents = read_file_contents(path);

	if (file_contents.empty())
		throw std::runtime_error(std::format("Level::Level({}): file contents are empty!", path.string()));

	auto [json, err] = json::parse(read_file_contents(path));

	if (err != "")
		throw std::runtime_error(std::format("Level::Level({}): file contains invalid JSON: {}", path.string(), err));

	Deserialize(json);
}

Level::Level(const json::value& json)
{
	Deserialize(json);
}

void Level::Deserialize(const json::value& json)
{
	name = json["name"].as_string();
	description = json["description"].as_string();
	author = json["author"].as_string();
	tileset_path = json["tileset-path"].as_string();

	auto layer_json_array = json["layers"].as_array();

	for (auto layer_json : layer_json_array)
	{
		layers.emplace_back(new Layer(layer_json));
	}

	auto entities_json_array = json["entities"].as_array();

	for (auto& entity_json : entities_json_array)
	{
		entities.emplace_back(new Entity(entity_json));
	}
}

json::value Level::Serialize() const
{
	json::object data;
	data["name"] = name;
	data["description"] = description;
	data["author"] = author;
	data["tileset-path"] = tileset_path;
	data["editor-version"] = REDACTED_EDITOR_LIB_VERSION;
	data["map-type"] = "OrthogonalTileMap";
	data["map-rows"] = (double)rows;
	data["map-cols"] = (double)cols;
	data["tile-width"] = 16.f;
	data["tile-height"] = 16.f;
	data["bgcolor"] = JsonConversions::deparse_color_to_hex(Colors::Black);
	data["tags"] = JsonConversions::deparse_string_list(tags);

	json::array layers_json;
	for (auto& layer : this->layers)
	{
		layers_json.push_back(layer->Serialize());
	}
	data["layers"] = layers_json;


	json::array entities_json;

	for (auto& entity : this->entities)
	{
		entities_json.push_back(entity->Serialize());
	}
	data["entities"] = entities_json;

	return data;
}
